/*
 *  Lucidor - e-book reader
 *  Copyright © 2013  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

var EXPORTED_SYMBOLS = [ "EpubCFI" ];

Components.utils.import("chrome://lucidor/content/utils.jsm");

/*
 * Create an EpubCFI object
 *
 * indirection_resolver: object with a function resolve
 *			 taking an indirection node as argument and
 *			 returning { href, node }
 */
function EpubCFI(indirection_resolver) {
    this.wrappedJSObject = this;
    this.path = [];
    this.type = "simple";
    this.filename = null;
    this.indirection_resolver = indirection_resolver;
}

EpubCFI.prototype = {
    CFI_PATH: "path",
    CFI_START: "start",
    CFI_END: "end",

    /*
     * Prepare EpubCFI object to be a range CFI
     */
    set_range: function() {
	this.start = [];
	this.end = [];
	this.type = "range";
    },

    /*
     * Set assertion on a step
     *
     * type: assertion type: "id", "text" (, "unclassified", "unknown")
     * value: array of csv values
     */
    set_assertion: function(step, type, value) {
	debug("set_assertion: " + type + " " + value[0]);
	if (!step.assertion)
	    step.assertion = { parameters: {} };
	step.assertion.type = type;
	step.assertion.value = value;
    },

    set_id_assertion: function(step, id) {
	this.set_assertion(step, "id", [id]);
    },

    set_text_assertion: function(step, before, after) {
	this.set_assertion(step, "text", [before, after]);
    },

    /*
     * Set assertion parameter on a step
     *
     * name: parameter name: "s"
     * value: array of csv values
     */
    set_parameter: function(step, name, value) {
	if (!step.assertion)
	    step.assertion = { parameters: {} };
	this.assertion.parameters[name] = value;
    },

    /*
     * Set sidebias parameter on a step
     *
     * value: "b" (before) or "a" (after)
     */
    set_sidebias_parameter: function(step, value) {
	this.set_parameter(step, "s", value);
    },

    /*
     * Prepend steps on a path
     *
     * path: EpubCFI.CFI_PATH, EpubCFI.CFI_START or EpubCFI.CFI_END
     * steps: array of steps
     */
    prepend_steps: function(path, steps) {
	this[path] = steps.concat(this[path]);
    },

    /*
     * Append steps on a path
     *
     * path: EpubCFI.CFI_PATH, EpubCFI.CFI_START or EpubCFI.CFI_END
     * steps: array of steps
     */
    append_steps: function(path, steps) {
	this[path] = this[path].concat(steps);
    },

    /*
     * Get a new step
     *
     * type: step type: "step", "indirectstep", "termstep"
     */
    get_step: function(type) {
	return { type: type,
		 assertion: { type: null,
			      value: null,
			      parameters: {} } };
    },

    /*
     * Get a new indirection step
     */
    get_indirection_step: function() {
	return this.get_step("indirectstep");
    },

    /*
     * Get a new character termination step
     *
     * offset: character offset
     */
    get_char_termination_step: function(offset) {
	let step = this.get_step("termstep");
	step.terminus = { type: ":",
			  offset: offset };
	return step;
    },

    /*
     * Get a new time termination step
     *
     * offset_time: time offset in seconds
     */
    get_time_termination_step: function(offset_time) {
	let step = this.get_step("termstep");
	step.terminus = { type: "~",
			  offset_time: offset_time,
			  offset_x: -1,
			  offset_y: -1 };
	return step;
    },

    /*
     * Get a new space termination step
     *
     * offset_x: x offset in percent (0 - 100)
     * offset_y: y offset in percent (0 - 100)
     */
    get_space_termination_step: function(offset_x, offset_y) {
	let step = this.get_step("termstep");
	step.terminus = { type: "@",
			  offset_time: -1,
			  offset_x: offset_x,
			  offset_y: offset_y };
	return step;
    },

    /*
     * Get a new time/space termination step
     *
     * offset_time: time offset in seconds
     * offset_x: x offset in percent (0 - 100)
     * offset_y: y offset in percent (0 - 100)
     */
    get_timespace_termination_step: function(offset_time, offset_x, offset_y) {
	let step = this.get_step("termstep");
	step.terminus = { type: "~@",
			  offset_time: offset_time,
			  offset_x: offset_x,
			  offset_y: offset_y };
	return step;
    },

    /*
     * Get a new before first element step
     */
    get_beforefirst_step: function() {
	debug("get_beforefirst_step");
	let step = this.get_step("step");
	step.offset = 1;
	return step;
    },

    /*
     * Get a new element step
     *
     * elem: element
     * after: whether to point after the element: true/false
     */
    get_element_step: function(elem, after) {
	debug("get_element_step");
	if (elem == null)
	    return this.get_beforefirst_step();

	let step = this.get_step("step");
	if (elem.hasAttribute("id"))
	    this.set_id_assertion(step, elem.getAttribute("id"));
	let offset = 0;
	for (let ch in loop(elem.parentNode.childNodes)) {
	    if (ch.nodeType == ch.ELEMENT_NODE)
		offset += 2;
	    if (ch == elem) {
		break;
	    }
	}
	if (after)
	    offset++;
	step.offset = offset;
	return step;
    },

    /*
     * For internal use.
     */
    get_element_steps: function(elem, after, rootnode) {
	debug("get_element_steps: " + elem.localName + " " +
	      rootnode.localName);
	if (elem == rootnode)
	    return [];
	let steps = [ this.get_element_step(elem, after) ];
	for (elem = elem.parentNode; elem != rootnode; elem = elem.parentNode)
	    steps.unshift(this.get_element_step(elem, false));
	return steps;
    },

    /*
     * For internal use.
     */
    is_luci_node: function(node) {
	return (node.nodeType == node.ELEMENT_NODE &&
		node.hasAttribute("class") &&
		node.getAttribute("class") == "lucitemp")
    },

    /*
     * For internal use.
     */
    get_text_steps: function(node, offset) {
	debug("get_text_steps");
	let off = offset;
	let steps = [];
	for (node = node.previousSibling; node; node = node.previousSibling) {
	    if (this.is_luci_node(node))
		continue;
	    if (node.nodeType == node.ELEMENT_NODE) {
		steps.push(this.get_element_step(node, true));
		break;
	    } else if (node.nodeType == node.TEXT_NODE) {
		off += node.nodeValue.length;
	    }
	}
	if (steps.length == 0)
	    steps.push(this.get_beforefirst_step());
	steps.push(this.get_char_termination_step(off));
	return steps;
    },

    /*
     * For internal use.
     */
    get_node_offset_steps: function(node, offset, rootnode) {
	debug("get_node_offset_steps");
	let steps;
	if (node.nodeType == node.ELEMENT_NODE) {
	    steps = this.get_element_steps(node, false, rootnode);
	    // Ignore offset
	} else {
	    steps = this.get_element_steps(node.parentNode, false, rootnode);
	    steps = steps.concat(this.get_text_steps(node, offset));
	}
	return steps;
    },

    /*
     * Create from element
     */
    create_from_element: function(elem, rootnode) {
	debug("create_from_element: " + elem.localName + " " +
	      rootnode.localName);
	let steps = this.get_element_steps(elem, false, rootnode);
	this.append_steps(this.CFI_PATH, steps);
    },

    /*
     * Create from DOMRange
     */
    create_from_domrange: function(range, rootnode) {
	debug("create_from_domrange");
	if (range.collapsed) {
	    let steps = this.get_node_offset_steps(range.startContainer,
						   range.startOffset,
						   rootnode);
	    this.append_steps(this.CFI_PATH, steps);
	} else {
	    let st = [];
	    for (let node = range.startContainer;
		 node != rootnode.parentNode;
		 node = node.parentNode)
		st.push(node);
	    let common;
	    for (common = range.endContainer;
		 common != rootnode.parentNode;
		 common = common.parentNode)
		if (st.indexOf(common) >= 0)
		    break;
	    if (common.nodeType != common.ELEMENT_NODE)
		common = common.parentNode;
	    this.set_range();

	    // Parent
	    let steps = this.get_element_steps(common, false, rootnode);
	    this.append_steps(this.CFI_PATH, steps);

	    // Start
	    steps = this.get_node_offset_steps(range.startContainer,
					       range.startOffset, common);
	    this.append_steps(this.CFI_START, steps);

	    // End
	    steps = this.get_node_offset_steps(range.endContainer,
					       range.endOffset, common);
	    this.append_steps(this.CFI_END, steps);
	}
    },

    /*
     * Convert to string
     */
    toString: function() {
	let pre = "";
	if (this.filename)
	    pre = this.filename + "#";
	let s = this.pathToString(this.path);
	if (this.type == "range") {
	    s += ",";
	    s += this.pathToString(this.start);
	    s += ",";
	    s += this.pathToString(this.end);
	}
	return pre + "epubcfi(" + s + ")";
    },

    /*
     * Convert path to string
     * For internal use
     *
     * cfipath: this.path, this.start or this.end
     */
    pathToString: function(cfipath) {
	let s = "";
	for (let step in loop(cfipath)) {
	    if (step.type == "step") {
		s += "/" + step.offset;
	    } else if (step.type == "indirectstep") {
		s += "!";
	    } else if (step.type == "termstep") {
		if (step.terminus.type == ":") {
		    s += ":" + step.terminus.offset;
		} else if (step.terminus.type == "~") {
		    s += "~" + step.terminus.offset_time;
		} else if (step.terminus.type == "@") {
		    s += "@" + step.terminus.offset_x +
			":" + step.terminus.offset_y;
		} else if (step.terminus.type == "~@") {
		    s += "~" + step.terminus.offset_time;
		    s += "@" + step.terminus.offset_x +
			":" + step.terminus.offset_y;
		}
	    }
	    if (step.assertion &&
		(step.assertion.value ||
		 (step.assertion.parameters &&
		  Object.keys(step.assertion.parameters).length > 0))) {
		s += "[";
		if (step.assertion.value) {
		    s += step.assertion.value[0];
		    for (let i = 1; i < step.assertion.value.length; i++)
			s += "," + step.assertion.value[i];
		}
		if (step.assertion.parameters) {
		    for (let name in step.assertion.parameters) {
			s += ";" + this.escape(name) + "=" +
			    this.escape(step.assertion.parameters[name]);
		    }
		}
		s += "]";
	    }
	}
	return s;
    },

    /*
     * Escape special characters in string 
     *
     * s: string to operate on
     */
    escape: function(s) {
	s = s.replace('^', '^^', 'g');
	s = s.replace('[', '^[', 'g');
	s = s.replace(']', '^]', 'g');
	s = s.replace('(', '^(', 'g');
	s = s.replace(')', '^)', 'g');
	s = s.replace(',', '^,', 'g');
	s = s.replace(';', '^;', 'g');
	s = s.replace('=', '^=', 'g');
	return s;
    },

    /*
     * Unescape special characters in string 
     *
     * s: string to operate on
     */
    /*
    unescape: function(s) {
	return s;
    },
    */

    /*
     * Resolve part
     *
     * returns href
     */
    resolve_part: function() {
	if (!this.indirection_resolver)
	    return null;

	let res = this.indirection_resolver.resolve(null);
	let pos = this.resolve_path(this.path, 0, res.node);
	if (pos.finished || !pos.indirection)
	    return null;
	return this.indirection_resolver.resolve(pos.indirection).href;
    },

    /*
     * Resolve
     *
     * returns DOMRange
     */
    resolve: function() {
	debug("resolve");
	if (!this.indirection_resolver)
	    return null;

	let pos = { finished: false,
		    nextstep: 0,
		    indirection: null };
	while (!pos.finished) {
	    let node = this.indirection_resolver.resolve(pos.indirection).node;
	    if (!node) {
		debug("error 2");
		return null;
	    }
	    pos = this.resolve_path(this.path, pos.nextstep, node);
	}
	if (this.type == "simple") {
	    if (pos.range)
		return pos.range;
	} else if (this.type == "range") {
	    if (!pos.node) {
		// error
		debug("error 4");
		return null;
	    }
	    let start = this.resolve_path(this.start, 0, pos.node);
	    let end = this.resolve_path(this.end, 0, pos.node);
	    if (!start.range || !end.range) {
		debug("error 6");
		return null;
	    }
	    let range = start.range;
	    range.setEnd(end.range.startContainer, end.range.startOffset);
	    return range;
	}
	debug("error 10");
	return null;
    },

    /*
     * Resolve path until next indirection
     * For internal use
     *
     * cfipath: this.path, this.start or this.end
     * nextstep: next step to process
     * rootnode: node to use as root 
     */
    resolve_path: function(cfipath, nextstep, rootnode) {
	let position = { parent: null,
			 node: null,
			 after: false,
			 range: null,
			 terminus: null,
			 finished: false };

	let parent = null;
	let node = rootnode;
	let after = false;
	while (nextstep < cfipath.length) {
	    let step = cfipath[nextstep];
	    nextstep++;
	    debug("step: " + step.type);

	    if (step.type == "step") {
		debug("step offset: " + step.offset);
		if (after) {
		    // Error
		    debug("epubcfi error 2");
		}

		//if (Math.mod(step.offset, 2) != 0)
		//	after = true;

		if (step.offset == 1) {
		    parent = node;
		    after = true;
		} else {
		    let i = 0;
		    for (let ch in loop(node.childNodes)) {
			if (ch.nodeType == ch.ELEMENT_NODE) {
			    i += 2;
			    if (step.offset == i) {
				node = ch;
				break;
			    }
			    if (step.offset == i + 1) {
				i++;
				node = ch;
				after = true;
				break;
			    }
			}
		    }
		    if (i < step.offset) {
			// error
			debug("epubcfi error 3");
		    }
		}
		if (step.assertion) {
		    if (step.assertion.type == "id") {
			if (!(node.hasAttribute("id") && step.assertion.value &&
			      step.assertion.value[0] == node.getAttribute("id"))) {
			    // FIXME
			    debug("Incorrect id assertion");
			}
		    }

		    if (step.assertion.parameters) {
			//if (step.assertion.parameters['vnd.']) {
			//}
		    }
		}
	    } else if (step.type == "indirectstep") {
		position.indirection = node;
		position.nextstep = nextstep;
		return position;
	    } else if (step.type == "termstep") {
		position.terminus = step.terminus;

		if (step.terminus.type == ":") {	// Character offset
		    debug(": " + step.terminus.offset);
		    if (!after) {
			debug("error: charterm (or inside img?)");
		    }
		    // FIXME
		    let range = rootnode.ownerDocument.createRange();
		    let pair;
		    if (parent) {
			pair = element_to_node_offset(parent,
						      step.terminus.offset);
		    } else {
			pair = after_element_offset(node,
						    step.terminus.offset);
		    }
		    range.setStart(pair.node, pair.offset);
		    position.range = range;
		} else if (step.terminus.type == "~") {	// Temporal offset
		    // Ignore for now
		} else if (step.terminus.type == "@") {	// Spatial offset
		    // Ignore for now
		} else if (step.terminus.type == "~@") {	// Temporal and spatial offset
		    // Ignore for now
		}
		if (step.assertion) {
		    if (step.assertion.type == "text") {
			if (step.terminus.type != ":") {
			    // error
			    debug("epubcfi error 5");
			}
			// Ignore for now
			// step.assertion.value[0]
			// step.assertion.value[1]
		    }

		    if (step.assertion.parameters['s']) {	// sidebias
			// Ignore for now
			// step.assertion.parameters['s']
		    }
		    //if (step.assertion.parameters['vnd.']) {
		    //}
		}
	    }
	}

	position.parent = parent;	// non-null iff position before first child
	position.node = node;
	position.after = after;
	position.finished = true;
	return position;
    },

    /*
     * Parse EPUB CFI string
     *
     * cfiexp: EPUB CFI string
     */
    parse: function(cfiexp) {
	try {
	    let cfi = pegparser.parse(cfiexp);
	    this.type = cfi.type;
	    this.path = cfi.path;
	    if (this.type == "range") {
		this.start = cfi.start;
		this.end = cfi.end;
	    }
	    this.classify_assertions();
	    return true;
	} catch (e) {
	    debug("parse error: " + e);
	}
	return false;
    },

    /*
     * Classify unclassified assertions
     * For internal use.
     */
    classify_assertions: function() {
	let paths;
	if (this.type == "range")
	    paths = [ this.path, this.start, this.end ];
	else
	    paths = [ this.path ];
	for (let path in loop(paths)) {
	    for (let step in loop(path)) {
		if (step.assertion && step.assertion.type == "unclassified") {
		    if (step.type == "termstep" && step.terminus.type == ":") {
			step.assertion.type = "text";
		    } else if (step.type == "step") {
			step.assertion.type = "id";
		    } else {
			step.assertion.type = "unknown";
		    }
		}
	    }
	}
    },

    /*
     * Compare position of this epubcfi relative to another epubcfi.
     *
     * returns -1 if this comes before other
     * returns 1 if the other comes before this
     * returns 0 if this and the other are equal
     */
    compare: function(other) {
	let aa = this.path;
	if (this.type == "range")
	    aa = aa.concat(this.start);
	let bb = other.path;
	if (other.type == "range")
	    bb = bb.concat(other.start);

	for (let i = 0; i < Math.min(aa.length, bb.length); i++) {
	    if (aa[i].type == "step") {
		if (bb[i].type == "step") {
		    if (aa[i].offset < bb[i].offset)
			return -1;
		    else if (aa[i].offset > bb[i].offset)
			return 1;
		    // continue;
		} else if (bb[i].type == "indirectstep") {
		    return 1;
		} else if (bb[i].type == "termstep") {
		    if (bb[i].terminus.type == ":")
			return -1;
		    else
			return 1;
		}
	    } else if (aa[i].type == "indirectstep") {
		if (bb[i].type == "step") {
		    return -1;
		} else if (bb[i].type == "indirectstep") {
		    // continue;
		} else if (bb[i].type == "termstep") {
		    return -1;
		}
	    } else if (aa[i].type == "termstep") {
		if (bb[i].type == "step") {
		    if (aa[i].terminus.type == ":")
			return 1;
		    else
			return -1;
		} else if (bb[i].type == "indirectstep") {
		    return 1;
		} else if (bb[i].type == "termstep") {
		    if (aa[i].terminus.type == ":") {
			if (bb[i].terminus.type == ":") {
			    if (aa[i].terminus.offset < bb[i].terminus.offset)
				return -1;
			    else if (aa[i].terminus.offset > bb[i].terminus.offset)
				return 1;
			    // continue;
			} else {
			    return 1;
			}
		    } else {
			// "temporal position is more important than spatial"
			if (aa[i].terminus.offset_time < bb[i].terminus.offset_time)
			    return -1;
			else if (aa[i].terminus.offset_time > bb[i].terminus.offset_time)
			    return 1;

			// "the y position is more important than x"
			if (aa[i].terminus.offset_y < bb[i].terminus.offset_y)
			    return -1;
			else if (aa[i].terminus.offset_y > bb[i].terminus.offset_y)
			    return 1;

			if (aa[i].terminus.offset_x < bb[i].terminus.offset_x)
			    return -1;
			else if (aa[i].terminus.offset_x > bb[i].terminus.offset_x)
			    return 1;
			// continue;
		    }
		}
	    }
	}
	if (aa.length < bb.length)
	    return -1;
	if (aa.length > bb.length)
	    return 1;
	return 0;
    }
}

function after_element_offset(element, offset) {
    let off = 0;
    let last = null;
    for (let node = element.nextSibling; node; node = node.nextSibling) {
	if (node.nodeType == node.TEXT_NODE) {
	    if (offset == 0) {
		return { node: node, offset: 0 };
	    } else if (off + node.nodeValue.length == offset) {
		// Offset points to the end of a text node
		return { node: node, offset: offset - off };
	    } else if (off + node.nodeValue.length > offset) {
		return { node: node, offset: offset - off };
	    } else {
		off += node.nodeValue.length;
	    }
	    last = node;
	} else if (node.nodeType == node.ELEMENT_NODE) {
	    break;
	}
	last = node;
    }
    // FIXME
    // This should not happen if offset > 0
    if (last)
	return { node: last, offset: last.nodeValue.length };
    return { node: element, offset: 0 };
    //return { node: element, offset: element.childNodes.length };
}

// Convert element+offset to (text)node+offset
function element_to_node_offset(element, offset) {
    //if (element.nodeType != element.ELEMENT_NODE)
    //	return { node: element, offset: offset };
    let off = 0;
    for (let node = element.firstChild; node; node = node.nextSibling) {
	if (node.nodeType == node.TEXT_NODE) {
	    if (offset == 0) {
		return { node: node, offset: 0 };
	    } else if (off + node.nodeValue.length == offset) {
		// Offset points to the end of a text node
		// Find beginning of next text node
		for (let n = node.nextSibling; n; n = n.nextSibling)
		    if (n.nodeType == n.TEXT_NODE)
			return { node: n, offset: 0 };
		// This should not happen
		return { node: node, offset: offset - off };
	    } else if (off + node.nodeValue.length > offset) {
		return { node: node, offset: offset - off };
	    } else {
		off += node.nodeValue.length;
	    }
	}
    }
    // This should not happen if offset > 0
    return { node: element, offset: offset };
    //return { node: element, offset: element.childNodes.length };
}

var pegparser = (function() {
  /*
   * Generated by PEG.js 0.7.0.
   *
   * http://pegjs.majda.cz/
   */

  function peg$subclass(child, parent) {
    function ctor() { this.constructor = child; }
    ctor.prototype = parent.prototype;
    child.prototype = new ctor();
  }

  function SyntaxError(expected, found, offset, line, column) {
    function buildMessage(expected, found) {
      function stringEscape(s) {
        function hex(ch) { return ch.charCodeAt(0).toString(16).toUpperCase(); }

        return s
          .replace(/\\/g,   '\\\\')
          .replace(/"/g,    '\\"')
          .replace(/\x08/g, '\\b')
          .replace(/\t/g,   '\\t')
          .replace(/\n/g,   '\\n')
          .replace(/\f/g,   '\\f')
          .replace(/\r/g,   '\\r')
          .replace(/[\x00-\x07\x0B\x0E\x0F]/g, function(ch) { return '\\x0' + hex(ch); })
          .replace(/[\x10-\x1F\x80-\xFF]/g,    function(ch) { return '\\x'  + hex(ch); })
          .replace(/[\u0180-\u0FFF]/g,         function(ch) { return '\\u0' + hex(ch); })
          .replace(/[\u1080-\uFFFF]/g,         function(ch) { return '\\u'  + hex(ch); });
      }

      var expectedDesc, foundDesc;

      switch (expected.length) {
        case 0:
          expectedDesc = "end of input";
          break;

        case 1:
          expectedDesc = expected[0];
          break;

        default:
          expectedDesc = expected.slice(0, -1).join(", ")
            + " or "
            + expected[expected.length - 1];
      }

      foundDesc = found ? "\"" + stringEscape(found) + "\"" : "end of input";

      return "Expected " + expectedDesc + " but " + foundDesc + " found.";
    }

    this.expected = expected;
    this.found    = found;
    this.offset   = offset;
    this.line     = line;
    this.column   = column;

    this.name     = "SyntaxError";
    this.message  = buildMessage(expected, found);
  }

  peg$subclass(SyntaxError, Error);

  function parse(input) {
    var options = arguments.length > 1 ? arguments[1] : {},

        peg$startRuleFunctions = { fragment: peg$parsefragment },
        peg$startRuleFunction  = peg$parsefragment,

        peg$c0 = null,
        peg$c1 = "epubcfi(",
        peg$c2 = "\"epubcfi(\"",
        peg$c3 = ")",
        peg$c4 = "\")\"",
        peg$c5 = function(r) { return r; },
        peg$c6 = function(p) { return { type: 'simple', path: p }; },
        peg$c7 = function(s, l) { return [s].concat(l); },
        peg$c8 = ",",
        peg$c9 = "\",\"",
        peg$c10 = function(p, s, e) { return { type: 'range', path: p, start: s, end: e }; },
        peg$c11 = [],
        peg$c12 = "",
        peg$c13 = function(s, t) { return t ? s.concat([t]) : s; },
        peg$c14 = "/",
        peg$c15 = "\"/\"",
        peg$c16 = function(n, a) { return { type: 'step', offset: n, assertion: a }; },
        peg$c17 = "!",
        peg$c18 = "\"!\"",
        peg$c19 = function() { return { type: 'indirectstep' }; },
        peg$c20 = function(t, a) { return { type: 'termstep', terminus: t, assertion: a }; },
        peg$c21 = ":",
        peg$c22 = "\":\"",
        peg$c23 = function(a) { return { type: ':', offset: a }; },
        peg$c24 = "~",
        peg$c25 = "\"~\"",
        peg$c26 = "@",
        peg$c27 = "\"@\"",
        peg$c28 = function(b, c, d) { return { type: '~@', offset_time: b, offset_x: c, offset_y: d }; },
        peg$c29 = function(e) { return { type: '~', offset_time: e, offset_x: -1, offset_y: -1 }; },
        peg$c30 = function(f, g) { return { type: '@', offset_time: -1, offset_x: f, offset_y: g }; },
        peg$c31 = function(a) { return parseFloat('0' + a, 10); },
        peg$c32 = function(b, c, d) { return parseFloat(b + c.join('') + d, 10); },
        peg$c33 = ".",
        peg$c34 = "\".\"",
        peg$c35 = function(a) { return '.' + a.join(''); },
        peg$c36 = function(z) { return parseInt(z, 10); },
        peg$c37 = function(a, b) { return parseInt(a + b.join(''), 10); },
        peg$c38 = "[",
        peg$c39 = "\"[\"",
        peg$c40 = "]",
        peg$c41 = "\"]\"",
        peg$c42 = function(v, p) { return { type: 'unclassified', value: v, parameters: p }; },
        peg$c43 = ";",
        peg$c44 = "\";\"",
        peg$c45 = "=",
        peg$c46 = "\"=\"",
        peg$c47 = function(n, v) { return { name: n, value: v }; },
        peg$c48 = function(a, b) { return [a].concat(b); },
        peg$c49 = function(c) { return [c]; },
        peg$c50 = function(a) { return a; },
        peg$c51 = function(a) { return a.join(''); },
        peg$c52 = function(b) { return b; },
        peg$c53 = function(c) { return c; },
        peg$c54 = function(d) { return d; },
        peg$c55 = function(e) { return e; },
        peg$c56 = function(f) { return f; },
        peg$c57 = "1",
        peg$c58 = "\"1\"",
        peg$c59 = "2",
        peg$c60 = "\"2\"",
        peg$c61 = "3",
        peg$c62 = "\"3\"",
        peg$c63 = "4",
        peg$c64 = "\"4\"",
        peg$c65 = "5",
        peg$c66 = "\"5\"",
        peg$c67 = "6",
        peg$c68 = "\"6\"",
        peg$c69 = "7",
        peg$c70 = "\"7\"",
        peg$c71 = "8",
        peg$c72 = "\"8\"",
        peg$c73 = "9",
        peg$c74 = "\"9\"",
        peg$c75 = "0",
        peg$c76 = "\"0\"",
        peg$c77 = " ",
        peg$c78 = "\" \"",
        peg$c79 = "^",
        peg$c80 = "\"^\"",
        peg$c81 = "\"",
        peg$c82 = "\"\\\"\"",
        peg$c83 = "(",
        peg$c84 = "\"(\"",
        peg$c85 = "any character",

        peg$currPos          = 0,
        peg$reportedPos      = 0,
        peg$cachedPos        = 0,
        peg$cachedPosDetails = { line: 1, column: 1, seenCR: false },
        peg$maxFailPos       = 0,
        peg$maxFailExpected  = [],
        peg$silentFails      = 0,

        peg$result;

    if ("startRule" in options) {
      if (!(options.startRule in peg$startRuleFunctions)) {
        throw new Error("Can't start parsing from rule \"" + options.startRule + "\".");
      }

      peg$startRuleFunction = peg$startRuleFunctions[options.startRule];
    }

    function text() {
      return input.substring(peg$reportedPos, peg$currPos);
    }

    function offset() {
      return peg$reportedPos;
    }

    function line() {
      return peg$computePosDetails(peg$reportedPos).line;
    }

    function column() {
      return peg$computePosDetails(peg$reportedPos).column;
    }

    function peg$computePosDetails(pos) {
      function advance(details, startPos, endPos) {
        var p, ch;

        for (p = startPos; p < endPos; p++) {
          ch = input.charAt(p);
          if (ch === "\n") {
            if (!details.seenCR) { details.line++; }
            details.column = 1;
            details.seenCR = false;
          } else if (ch === "\r" || ch === "\u2028" || ch === "\u2029") {
            details.line++;
            details.column = 1;
            details.seenCR = true;
          } else {
            details.column++;
            details.seenCR = false;
          }
        }
      }

      if (peg$cachedPos !== pos) {
        if (peg$cachedPos > pos) {
          peg$cachedPos = 0;
          peg$cachedPosDetails = { line: 1, column: 1, seenCR: false };
        }
        advance(peg$cachedPosDetails, peg$cachedPos, pos);
        peg$cachedPos = pos;
      }

      return peg$cachedPosDetails;
    }

    function peg$fail(expected) {
      if (peg$currPos < peg$maxFailPos) { return; }

      if (peg$currPos > peg$maxFailPos) {
        peg$maxFailPos = peg$currPos;
        peg$maxFailExpected = [];
      }

      peg$maxFailExpected.push(expected);
    }

    function peg$cleanupExpected(expected) {
      var i = 0;

      expected.sort();

      while (i < expected.length) {
        if (expected[i - 1] === expected[i]) {
          expected.splice(i, 1);
        } else {
          i++;
        }
      }
    }

    function peg$parsefragment() {
      var s0, s1, s2, s3;

      s0 = peg$currPos;
      if (input.substr(peg$currPos, 8) === peg$c1) {
        s1 = peg$c1;
        peg$currPos += 8;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c2); }
      }
      if (s1 !== null) {
        s2 = peg$parserange();
        if (s2 !== null) {
          if (input.charCodeAt(peg$currPos) === 41) {
            s3 = peg$c3;
            peg$currPos++;
          } else {
            s3 = null;
            if (peg$silentFails === 0) { peg$fail(peg$c4); }
          }
          if (s3 !== null) {
            peg$reportedPos = s0;
            s1 = peg$c5(s2);
            if (s1 === null) {
              peg$currPos = s0;
              s0 = s1;
            } else {
              s0 = s1;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }
      if (s0 === null) {
        s0 = peg$currPos;
        if (input.substr(peg$currPos, 8) === peg$c1) {
          s1 = peg$c1;
          peg$currPos += 8;
        } else {
          s1 = null;
          if (peg$silentFails === 0) { peg$fail(peg$c2); }
        }
        if (s1 !== null) {
          s2 = peg$parsepath();
          if (s2 !== null) {
            if (input.charCodeAt(peg$currPos) === 41) {
              s3 = peg$c3;
              peg$currPos++;
            } else {
              s3 = null;
              if (peg$silentFails === 0) { peg$fail(peg$c4); }
            }
            if (s3 !== null) {
              peg$reportedPos = s0;
              s1 = peg$c6(s2);
              if (s1 === null) {
                peg$currPos = s0;
                s0 = s1;
              } else {
                s0 = s1;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      }

      return s0;
    }

    function peg$parsepath() {
      var s0, s1, s2;

      s0 = peg$currPos;
      s1 = peg$parsestep();
      if (s1 !== null) {
        s2 = peg$parselocal_path();
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c7(s1,s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parserange() {
      var s0, s1, s2, s3, s4, s5;

      s0 = peg$currPos;
      s1 = peg$parsepath();
      if (s1 !== null) {
        if (input.charCodeAt(peg$currPos) === 44) {
          s2 = peg$c8;
          peg$currPos++;
        } else {
          s2 = null;
          if (peg$silentFails === 0) { peg$fail(peg$c9); }
        }
        if (s2 !== null) {
          s3 = peg$parselocal_path();
          if (s3 !== null) {
            if (input.charCodeAt(peg$currPos) === 44) {
              s4 = peg$c8;
              peg$currPos++;
            } else {
              s4 = null;
              if (peg$silentFails === 0) { peg$fail(peg$c9); }
            }
            if (s4 !== null) {
              s5 = peg$parselocal_path();
              if (s5 !== null) {
                peg$reportedPos = s0;
                s1 = peg$c10(s1,s3,s5);
                if (s1 === null) {
                  peg$currPos = s0;
                  s0 = s1;
                } else {
                  s0 = s1;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$c0;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parselocal_path() {
      var s0, s1, s2;

      s0 = peg$currPos;
      s1 = [];
      s2 = peg$parsestep();
      if (s2 === null) {
        s2 = peg$parseindirectstep();
      }
      while (s2 !== null) {
        s1.push(s2);
        s2 = peg$parsestep();
        if (s2 === null) {
          s2 = peg$parseindirectstep();
        }
      }
      if (s1 !== null) {
        s2 = peg$parsetermstep();
        if (s2 === null) {
          s2 = peg$c12;
        }
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c13(s1,s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parsestep() {
      var s0, s1, s2, s3;

      s0 = peg$currPos;
      if (input.charCodeAt(peg$currPos) === 47) {
        s1 = peg$c14;
        peg$currPos++;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c15); }
      }
      if (s1 !== null) {
        s2 = peg$parseinteger();
        if (s2 !== null) {
          s3 = peg$parseassertion();
          if (s3 === null) {
            s3 = peg$c12;
          }
          if (s3 !== null) {
            peg$reportedPos = s0;
            s1 = peg$c16(s2,s3);
            if (s1 === null) {
              peg$currPos = s0;
              s0 = s1;
            } else {
              s0 = s1;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parseindirectstep() {
      var s0, s1;

      s0 = peg$currPos;
      if (input.charCodeAt(peg$currPos) === 33) {
        s1 = peg$c17;
        peg$currPos++;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c18); }
      }
      if (s1 !== null) {
        peg$reportedPos = s0;
        s1 = peg$c19();
      }
      if (s1 === null) {
        peg$currPos = s0;
        s0 = s1;
      } else {
        s0 = s1;
      }

      return s0;
    }

    function peg$parsetermstep() {
      var s0, s1, s2;

      s0 = peg$currPos;
      s1 = peg$parseterminus();
      if (s1 !== null) {
        s2 = peg$parseassertion();
        if (s2 === null) {
          s2 = peg$c12;
        }
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c20(s1,s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parseterminus() {
      var s0, s1, s2, s3, s4, s5, s6;

      s0 = peg$currPos;
      if (input.charCodeAt(peg$currPos) === 58) {
        s1 = peg$c21;
        peg$currPos++;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c22); }
      }
      if (s1 !== null) {
        s2 = peg$parseinteger();
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c23(s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }
      if (s0 === null) {
        s0 = peg$currPos;
        if (input.charCodeAt(peg$currPos) === 126) {
          s1 = peg$c24;
          peg$currPos++;
        } else {
          s1 = null;
          if (peg$silentFails === 0) { peg$fail(peg$c25); }
        }
        if (s1 !== null) {
          s2 = peg$parsenumber();
          if (s2 !== null) {
            if (input.charCodeAt(peg$currPos) === 64) {
              s3 = peg$c26;
              peg$currPos++;
            } else {
              s3 = null;
              if (peg$silentFails === 0) { peg$fail(peg$c27); }
            }
            if (s3 !== null) {
              s4 = peg$parsenumber();
              if (s4 !== null) {
                if (input.charCodeAt(peg$currPos) === 58) {
                  s5 = peg$c21;
                  peg$currPos++;
                } else {
                  s5 = null;
                  if (peg$silentFails === 0) { peg$fail(peg$c22); }
                }
                if (s5 !== null) {
                  s6 = peg$parsenumber();
                  if (s6 !== null) {
                    peg$reportedPos = s0;
                    s1 = peg$c28(s2,s4,s6);
                    if (s1 === null) {
                      peg$currPos = s0;
                      s0 = s1;
                    } else {
                      s0 = s1;
                    }
                  } else {
                    peg$currPos = s0;
                    s0 = peg$c0;
                  }
                } else {
                  peg$currPos = s0;
                  s0 = peg$c0;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$c0;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
        if (s0 === null) {
          s0 = peg$currPos;
          if (input.charCodeAt(peg$currPos) === 126) {
            s1 = peg$c24;
            peg$currPos++;
          } else {
            s1 = null;
            if (peg$silentFails === 0) { peg$fail(peg$c25); }
          }
          if (s1 !== null) {
            s2 = peg$parsenumber();
            if (s2 !== null) {
              peg$reportedPos = s0;
              s1 = peg$c29(s2);
              if (s1 === null) {
                peg$currPos = s0;
                s0 = s1;
              } else {
                s0 = s1;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
          if (s0 === null) {
            s0 = peg$currPos;
            if (input.charCodeAt(peg$currPos) === 64) {
              s1 = peg$c26;
              peg$currPos++;
            } else {
              s1 = null;
              if (peg$silentFails === 0) { peg$fail(peg$c27); }
            }
            if (s1 !== null) {
              s2 = peg$parsenumber();
              if (s2 !== null) {
                if (input.charCodeAt(peg$currPos) === 58) {
                  s3 = peg$c21;
                  peg$currPos++;
                } else {
                  s3 = null;
                  if (peg$silentFails === 0) { peg$fail(peg$c22); }
                }
                if (s3 !== null) {
                  s4 = peg$parsenumber();
                  if (s4 !== null) {
                    peg$reportedPos = s0;
                    s1 = peg$c30(s2,s4);
                    if (s1 === null) {
                      peg$currPos = s0;
                      s0 = s1;
                    } else {
                      s0 = s1;
                    }
                  } else {
                    peg$currPos = s0;
                    s0 = peg$c0;
                  }
                } else {
                  peg$currPos = s0;
                  s0 = peg$c0;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$c0;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          }
        }
      }

      return s0;
    }

    function peg$parsenumber() {
      var s0, s1, s2, s3;

      s0 = peg$currPos;
      s1 = peg$parsezero();
      if (s1 !== null) {
        s2 = peg$parsedecimals();
        if (s2 === null) {
          s2 = peg$c12;
        }
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c31(s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }
      if (s0 === null) {
        s0 = peg$currPos;
        s1 = peg$parsedigit_non_zero();
        if (s1 !== null) {
          s2 = [];
          s3 = peg$parsedigit();
          while (s3 !== null) {
            s2.push(s3);
            s3 = peg$parsedigit();
          }
          if (s2 !== null) {
            s3 = peg$parsedecimals();
            if (s3 === null) {
              s3 = peg$c12;
            }
            if (s3 !== null) {
              peg$reportedPos = s0;
              s1 = peg$c32(s1,s2,s3);
              if (s1 === null) {
                peg$currPos = s0;
                s0 = s1;
              } else {
                s0 = s1;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      }

      return s0;
    }

    function peg$parsedecimals() {
      var s0, s1, s2, s3;

      s0 = peg$currPos;
      if (input.charCodeAt(peg$currPos) === 46) {
        s1 = peg$c33;
        peg$currPos++;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c34); }
      }
      if (s1 !== null) {
        s2 = [];
        s3 = peg$parsedigit();
        if (s3 !== null) {
          while (s3 !== null) {
            s2.push(s3);
            s3 = peg$parsedigit();
          }
        } else {
          s2 = peg$c0;
        }
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c35(s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parseinteger() {
      var s0, s1, s2, s3;

      s0 = peg$currPos;
      s1 = peg$parsezero();
      if (s1 !== null) {
        peg$reportedPos = s0;
        s1 = peg$c36(s1);
      }
      if (s1 === null) {
        peg$currPos = s0;
        s0 = s1;
      } else {
        s0 = s1;
      }
      if (s0 === null) {
        s0 = peg$currPos;
        s1 = peg$parsedigit_non_zero();
        if (s1 !== null) {
          s2 = [];
          s3 = peg$parsedigit();
          while (s3 !== null) {
            s2.push(s3);
            s3 = peg$parsedigit();
          }
          if (s2 !== null) {
            peg$reportedPos = s0;
            s1 = peg$c37(s1,s2);
            if (s1 === null) {
              peg$currPos = s0;
              s0 = s1;
            } else {
              s0 = s1;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      }

      return s0;
    }

    function peg$parseassertion() {
      var s0, s1, s2, s3, s4;

      s0 = peg$currPos;
      if (input.charCodeAt(peg$currPos) === 91) {
        s1 = peg$c38;
        peg$currPos++;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c39); }
      }
      if (s1 !== null) {
        s2 = peg$parsecsv();
        if (s2 === null) {
          s2 = peg$c12;
        }
        if (s2 !== null) {
          s3 = [];
          s4 = peg$parseparameter();
          while (s4 !== null) {
            s3.push(s4);
            s4 = peg$parseparameter();
          }
          if (s3 !== null) {
            if (input.charCodeAt(peg$currPos) === 93) {
              s4 = peg$c40;
              peg$currPos++;
            } else {
              s4 = null;
              if (peg$silentFails === 0) { peg$fail(peg$c41); }
            }
            if (s4 !== null) {
              peg$reportedPos = s0;
              s1 = peg$c42(s2,s3);
              if (s1 === null) {
                peg$currPos = s0;
                s0 = s1;
              } else {
                s0 = s1;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parseparameter() {
      var s0, s1, s2, s3, s4;

      s0 = peg$currPos;
      if (input.charCodeAt(peg$currPos) === 59) {
        s1 = peg$c43;
        peg$currPos++;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c44); }
      }
      if (s1 !== null) {
        s2 = peg$parsevalue_no_space();
        if (s2 !== null) {
          if (input.charCodeAt(peg$currPos) === 61) {
            s3 = peg$c45;
            peg$currPos++;
          } else {
            s3 = null;
            if (peg$silentFails === 0) { peg$fail(peg$c46); }
          }
          if (s3 !== null) {
            s4 = peg$parsecsv();
            if (s4 !== null) {
              peg$reportedPos = s0;
              s1 = peg$c47(s2,s4);
              if (s1 === null) {
                peg$currPos = s0;
                s0 = s1;
              } else {
                s0 = s1;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parsecsv() {
      var s0, s1, s2, s3;

      s0 = peg$currPos;
      s1 = peg$parsevalue();
      if (s1 === null) {
        s1 = peg$c12;
      }
      if (s1 !== null) {
        s2 = [];
        s3 = peg$parsecomma_value();
        if (s3 !== null) {
          while (s3 !== null) {
            s2.push(s3);
            s3 = peg$parsecomma_value();
          }
        } else {
          s2 = peg$c0;
        }
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c48(s1,s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }
      if (s0 === null) {
        s0 = peg$currPos;
        s1 = peg$parsevalue();
        if (s1 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c49(s1);
        }
        if (s1 === null) {
          peg$currPos = s0;
          s0 = s1;
        } else {
          s0 = s1;
        }
      }

      return s0;
    }

    function peg$parsecomma_value() {
      var s0, s1, s2;

      s0 = peg$currPos;
      if (input.charCodeAt(peg$currPos) === 44) {
        s1 = peg$c8;
        peg$currPos++;
      } else {
        s1 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c9); }
      }
      if (s1 !== null) {
        s2 = peg$parsevalue();
        if (s2 === null) {
          s2 = peg$c12;
        }
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c50(s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }

      return s0;
    }

    function peg$parsevalue() {
      var s0, s1, s2;

      s0 = peg$currPos;
      s1 = [];
      s2 = peg$parsecharacter_escaped_special();
      if (s2 !== null) {
        while (s2 !== null) {
          s1.push(s2);
          s2 = peg$parsecharacter_escaped_special();
        }
      } else {
        s1 = peg$c0;
      }
      if (s1 !== null) {
        peg$reportedPos = s0;
        s1 = peg$c51(s1);
      }
      if (s1 === null) {
        peg$currPos = s0;
        s0 = s1;
      } else {
        s0 = s1;
      }

      return s0;
    }

    function peg$parsevalue_no_space() {
      var s0, s1, s2;

      s0 = peg$currPos;
      s1 = [];
      s2 = peg$parsecharacter_escaped_special_no_space();
      if (s2 !== null) {
        while (s2 !== null) {
          s1.push(s2);
          s2 = peg$parsecharacter_escaped_special_no_space();
        }
      } else {
        s1 = peg$c0;
      }
      if (s1 !== null) {
        peg$reportedPos = s0;
        s1 = peg$c51(s1);
      }
      if (s1 === null) {
        peg$currPos = s0;
        s0 = s1;
      } else {
        s0 = s1;
      }

      return s0;
    }

    function peg$parsespecial_chars() {
      var s0;

      s0 = peg$parsecircumflex();
      if (s0 === null) {
        s0 = peg$parsesquare_brackets();
        if (s0 === null) {
          s0 = peg$parseparentheses();
          if (s0 === null) {
            s0 = peg$parsecomma();
            if (s0 === null) {
              s0 = peg$parsesemicolon();
              if (s0 === null) {
                s0 = peg$parseequal();
              }
            }
          }
        }
      }

      return s0;
    }

    function peg$parseescaped_special_chars() {
      var s0, s1, s2;

      s0 = peg$currPos;
      s1 = peg$parsecircumflex();
      if (s1 !== null) {
        s2 = peg$parsecircumflex();
        if (s2 !== null) {
          peg$reportedPos = s0;
          s1 = peg$c50(s2);
          if (s1 === null) {
            peg$currPos = s0;
            s0 = s1;
          } else {
            s0 = s1;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$c0;
      }
      if (s0 === null) {
        s0 = peg$currPos;
        s1 = peg$parsecircumflex();
        if (s1 !== null) {
          s2 = peg$parsesquare_brackets();
          if (s2 !== null) {
            peg$reportedPos = s0;
            s1 = peg$c52(s2);
            if (s1 === null) {
              peg$currPos = s0;
              s0 = s1;
            } else {
              s0 = s1;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
        if (s0 === null) {
          s0 = peg$currPos;
          s1 = peg$parsecircumflex();
          if (s1 !== null) {
            s2 = peg$parseparentheses();
            if (s2 !== null) {
              peg$reportedPos = s0;
              s1 = peg$c53(s2);
              if (s1 === null) {
                peg$currPos = s0;
                s0 = s1;
              } else {
                s0 = s1;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
          if (s0 === null) {
            s0 = peg$currPos;
            s1 = peg$parsecircumflex();
            if (s1 !== null) {
              s2 = peg$parsecomma();
              if (s2 !== null) {
                peg$reportedPos = s0;
                s1 = peg$c54(s2);
                if (s1 === null) {
                  peg$currPos = s0;
                  s0 = s1;
                } else {
                  s0 = s1;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$c0;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$c0;
            }
            if (s0 === null) {
              s0 = peg$currPos;
              s1 = peg$parsecircumflex();
              if (s1 !== null) {
                s2 = peg$parsesemicolon();
                if (s2 !== null) {
                  peg$reportedPos = s0;
                  s1 = peg$c55(s2);
                  if (s1 === null) {
                    peg$currPos = s0;
                    s0 = s1;
                  } else {
                    s0 = s1;
                  }
                } else {
                  peg$currPos = s0;
                  s0 = peg$c0;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$c0;
              }
              if (s0 === null) {
                s0 = peg$currPos;
                s1 = peg$parsecircumflex();
                if (s1 !== null) {
                  s2 = peg$parseequal();
                  if (s2 !== null) {
                    peg$reportedPos = s0;
                    s1 = peg$c56(s2);
                    if (s1 === null) {
                      peg$currPos = s0;
                      s0 = s1;
                    } else {
                      s0 = s1;
                    }
                  } else {
                    peg$currPos = s0;
                    s0 = peg$c0;
                  }
                } else {
                  peg$currPos = s0;
                  s0 = peg$c0;
                }
              }
            }
          }
        }
      }

      return s0;
    }

    function peg$parsecharacter_escaped_special_no_space() {
      var s0, s1, s2;

      s0 = peg$currPos;
      s1 = peg$parseescaped_special_chars();
      if (s1 !== null) {
        peg$reportedPos = s0;
        s1 = peg$c50(s1);
      }
      if (s1 === null) {
        peg$currPos = s0;
        s0 = s1;
      } else {
        s0 = s1;
      }
      if (s0 === null) {
        s0 = peg$currPos;
        s1 = peg$currPos;
        peg$silentFails++;
        s2 = peg$parsespecial_chars();
        if (s2 === null) {
          s2 = peg$parsespace();
        }
        peg$silentFails--;
        if (s2 === null) {
          s1 = peg$c12;
        } else {
          peg$currPos = s1;
          s1 = peg$c0;
        }
        if (s1 !== null) {
          s2 = peg$parsecharacter();
          if (s2 !== null) {
            peg$reportedPos = s0;
            s1 = peg$c52(s2);
            if (s1 === null) {
              peg$currPos = s0;
              s0 = s1;
            } else {
              s0 = s1;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$c0;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$c0;
        }
      }

      return s0;
    }

    function peg$parsecharacter_escaped_special() {
      var s0;

      s0 = peg$parsespace();
      if (s0 === null) {
        s0 = peg$parsecharacter_escaped_special_no_space();
      }

      return s0;
    }

    function peg$parsedigit() {
      var s0;

      s0 = peg$parsezero();
      if (s0 === null) {
        s0 = peg$parsedigit_non_zero();
      }

      return s0;
    }

    function peg$parsedigit_non_zero() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 49) {
        s0 = peg$c57;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c58); }
      }
      if (s0 === null) {
        if (input.charCodeAt(peg$currPos) === 50) {
          s0 = peg$c59;
          peg$currPos++;
        } else {
          s0 = null;
          if (peg$silentFails === 0) { peg$fail(peg$c60); }
        }
        if (s0 === null) {
          if (input.charCodeAt(peg$currPos) === 51) {
            s0 = peg$c61;
            peg$currPos++;
          } else {
            s0 = null;
            if (peg$silentFails === 0) { peg$fail(peg$c62); }
          }
          if (s0 === null) {
            if (input.charCodeAt(peg$currPos) === 52) {
              s0 = peg$c63;
              peg$currPos++;
            } else {
              s0 = null;
              if (peg$silentFails === 0) { peg$fail(peg$c64); }
            }
            if (s0 === null) {
              if (input.charCodeAt(peg$currPos) === 53) {
                s0 = peg$c65;
                peg$currPos++;
              } else {
                s0 = null;
                if (peg$silentFails === 0) { peg$fail(peg$c66); }
              }
              if (s0 === null) {
                if (input.charCodeAt(peg$currPos) === 54) {
                  s0 = peg$c67;
                  peg$currPos++;
                } else {
                  s0 = null;
                  if (peg$silentFails === 0) { peg$fail(peg$c68); }
                }
                if (s0 === null) {
                  if (input.charCodeAt(peg$currPos) === 55) {
                    s0 = peg$c69;
                    peg$currPos++;
                  } else {
                    s0 = null;
                    if (peg$silentFails === 0) { peg$fail(peg$c70); }
                  }
                  if (s0 === null) {
                    if (input.charCodeAt(peg$currPos) === 56) {
                      s0 = peg$c71;
                      peg$currPos++;
                    } else {
                      s0 = null;
                      if (peg$silentFails === 0) { peg$fail(peg$c72); }
                    }
                    if (s0 === null) {
                      if (input.charCodeAt(peg$currPos) === 57) {
                        s0 = peg$c73;
                        peg$currPos++;
                      } else {
                        s0 = null;
                        if (peg$silentFails === 0) { peg$fail(peg$c74); }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }

      return s0;
    }

    function peg$parsezero() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 48) {
        s0 = peg$c75;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c76); }
      }

      return s0;
    }

    function peg$parsespace() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 32) {
        s0 = peg$c77;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c78); }
      }

      return s0;
    }

    function peg$parsecircumflex() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 94) {
        s0 = peg$c79;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c80); }
      }

      return s0;
    }

    function peg$parsedouble_quote() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 34) {
        s0 = peg$c81;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c82); }
      }

      return s0;
    }

    function peg$parsesquare_brackets() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 91) {
        s0 = peg$c38;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c39); }
      }
      if (s0 === null) {
        if (input.charCodeAt(peg$currPos) === 93) {
          s0 = peg$c40;
          peg$currPos++;
        } else {
          s0 = null;
          if (peg$silentFails === 0) { peg$fail(peg$c41); }
        }
      }

      return s0;
    }

    function peg$parseparentheses() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 40) {
        s0 = peg$c83;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c84); }
      }
      if (s0 === null) {
        if (input.charCodeAt(peg$currPos) === 41) {
          s0 = peg$c3;
          peg$currPos++;
        } else {
          s0 = null;
          if (peg$silentFails === 0) { peg$fail(peg$c4); }
        }
      }

      return s0;
    }

    function peg$parsecomma() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 44) {
        s0 = peg$c8;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c9); }
      }

      return s0;
    }

    function peg$parsesemicolon() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 59) {
        s0 = peg$c43;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c44); }
      }

      return s0;
    }

    function peg$parseequal() {
      var s0;

      if (input.charCodeAt(peg$currPos) === 61) {
        s0 = peg$c45;
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c46); }
      }

      return s0;
    }

    function peg$parsecharacter() {
      var s0;

      if (input.length > peg$currPos) {
        s0 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s0 = null;
        if (peg$silentFails === 0) { peg$fail(peg$c85); }
      }

      return s0;
    }

    peg$result = peg$startRuleFunction();

    if (peg$result !== null && peg$currPos === input.length) {
      return peg$result;
    } else {
      peg$cleanupExpected(peg$maxFailExpected);
      peg$reportedPos = Math.max(peg$currPos, peg$maxFailPos);

      throw new SyntaxError(
        peg$maxFailExpected,
        peg$reportedPos < input.length ? input.charAt(peg$reportedPos) : null,
        peg$reportedPos,
        peg$computePosDetails(peg$reportedPos).line,
        peg$computePosDetails(peg$reportedPos).column
      );
    }
  }

  return {
    SyntaxError: SyntaxError,
    parse      : parse
  };
})();
