/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2014  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

const CNT_NS = "urn:oasis:names:tc:opendocument:xmlns:container";
const DC_NS = "http://purl.org/dc/elements/1.1/";
const OPF_NS = "http://www.idpf.org/2007/opf";
const XHTML_NS = "http://www.w3.org/1999/xhtml";

function epubCreator() {
    this.wrappedJSObject = this;
    this.title = 'Untitled';
    this.creators = new Array();
    this.languages = new Array();
    this.source = null;
    this.subjects = new Array();
    this.subtitle = null;
    this.publisher = null;
    this.modified = null;
    this.uid = null;
    this.spine = null;
    this.nonspine = null;
    this.nav = null;
    this.autopubs = new Array();
}

epubCreator.prototype = {
    classDescription: "epubCreator",
    classID: Components.ID("{73ce0ef7-2ae0-47eb-82ae-e63a256b499b}"),
    contractID: "@lucidor.org/epubCreator;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    set_title: function(title) {
	this.title = title;
    },

    set_publisher: function(publisher) {
	this.publisher = publisher;
    },

    set_uid: function(uid) {
	if (!uid) {
	    let uuid = Components.classes["@mozilla.org/uuid-generator;1"].
		getService(Components.interfaces.nsIUUIDGenerator).
		generateUUID().toString();
	    uid = "urn:uuid:" + uuid.substring(1, 37);
	}
	this.uid = uid;
    },

    add_autopub: function(type, uri) {
	this.autopubs.push({type: type, uri: uri});
    },

    set_modified: function(mod) {
	if (!mod)
	    mod = iso_date(new Date());
	this.modified = mod;
    },

    add_language: function(lang) {
	for (let i = 0; i < this.languages.length; i++)
	    if (this.languages[i] == lang)
		return;
	this.languages.push(lang);
    },

    add_subject: function(sub) {
	for (let i = 0; i < this.subjects.length; i++)
	    if (this.subjects[i] == sub)
		return;
	this.subjects.push(sub);
    },

    set_source: function(src) {
	this.source = src;
    },

    add_creator: function(name, role) {
	if (!role)
	    role = 'aut';
	this.creators.push({ name: name, role: role});
    },

    add_items: function(d_spine, d_nonspine, d_nav) {
	this.spine = d_spine;
	this.nonspine = d_nonspine;
	this.nav = d_nav;
    },

    generate: function() {
	debug("gen: 0");

	// check
	if (!this.uid)
	    this.set_uid(null);
	if (!this.modified)
	    this.set_modified(null);
	if (this.languages.length == 0)
	    this.add_language('und');

	// package
	let metadata = "";
	metadata += '<dc:identifier id="uid">' + escapeHtml(this.uid) + '</dc:identifier>\n';
	metadata += '<dc:title>' + escapeHtml(this.title) + '</dc:title>\n';
	for (let aut in loop(this.creators)) {
	    metadata += '<dc:creator>' + escapeHtml(aut.name) + '</dc:creator>\n';
	}
	for (let lang in loop(this.languages))
	    metadata += '<dc:language>' + escapeHtml(lang) + '</dc:language>\n';

	if (this.publisher)
	    metadata += '<dc:publisher>' + escapeHtml(this.publisher) + '</dc:publisher>\n';

	if (this.source)
	    metadata += '<dc:source>' + escapeHtml(this.source) + '</dc:source>\n';

	metadata += '<meta property="dcterms:modified">' + this.modified + '</meta>\n';

	for (let src in loop(this.autopubs))
	    metadata += '<meta property="luci:autopublication">' + src.type + ':' + escapeHtml(src.uri) + '</meta>\n';

	let manifest = '<item id="nav" href="nav.html" media-type="application/xhtml+xml" properties="nav"/>\n';

	debug("gen: 2");

	let id_counter = 0;
	function get_id() {
	    return "id" + id_counter++;
	}

	let spine = "";
	for (let it in loop(this.spine)) {
	    let sid = get_id();
	    manifest += '<item id="' + sid + '" href="' + escapeHtml(it.name) + '" media-type="' + escapeHtml(it.mediatype) + '"/>\n';

	    spine += '<itemref idref="' + sid + '"' 
	    if (it.nonlinear)
		spine += ' linear = "false"';
	    spine += '/>\n';
	}

	debug("init: 3");

	for (let it in loop(this.nonspine)) {
	    let sid = get_id();
	    manifest += '<item id="' + sid + '" href="' + escapeHtml(it.name) + '" media-type="' + escapeHtml(it.mediatype) + '"/>\n';
	}

	let pkg = '<?xml version="1.0"?>\n';
	pkg += '<package version="3.0" xmlns="http://www.idpf.org/2007/opf" unique-identifier="uid">\n';
	pkg += '<metadata xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:opf="http://www.idpf.org/2007/opf" xmlns:luci="http://lucidor.org/-/x-opf/">\n';
	pkg += metadata;
	pkg += '</metadata>\n';
	pkg += '<manifest>\n';
	pkg += manifest;
	pkg += '</manifest>\n';
	pkg += '<spine>\n';
	pkg += spine;
	pkg += '</spine>\n';
	pkg += '</package>\n';

	debug("init: 4");

	let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	    getService().wrappedJSObject;
	let zip_file = bookman.getTempdir();
	zip_file.append("lucidor.epub");
	debug(":" + zip_file.path);
	zip_file.createUnique(zip_file.NORMAL_FILE_TYPE, parseInt("0600", 8));

	let zw = Components.classes["@mozilla.org/zipwriter;1"].
		createInstance(Components.interfaces.nsIZipWriter);
	zw.open(zip_file, 0x02 | 0x20);

	let modtime = 0;

	debug("init: 6");

	let converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
	converter.charset = "UTF-8";

	let mimetype = "application/epub+zip";
	let stream = converter.convertToInputStream(mimetype);
	zw.addEntryStream("mimetype", modtime, zw.COMPRESSION_NONE,
			  stream, false);

	let container = '<?xml version="1.0"?>\n';
	container += '<container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">\n';
	container += '<rootfiles>\n';
	container += '<rootfile full-path="OPS/luci.opf" media-type="application/oebps-package+xml"/>\n';
	container += '</rootfiles>\n';
	container += '</container>';
	stream = converter.convertToInputStream(container);
	zw.addEntryStream("META-INF/container.xml", modtime,
			  zw.COMPRESSION_DEFAULT, stream, false);

	stream = converter.convertToInputStream(pkg);
	zw.addEntryStream("OPS/luci.opf", modtime,
			  zw.COMPRESSION_DEFAULT, stream, false);

	stream = converter.convertToInputStream(this.generate_nav());
	zw.addEntryStream("OPS/nav.html", modtime,
			  zw.COMPRESSION_DEFAULT, stream, false);

	debug("init: 9");

	for (let it in loop(this.spine)) {
	    let my_modtime = modtime;
	    if (it.string)
		stream = converter.convertToInputStream(it.string);
	    else	// href
		stream = hrefToStream(it.href);
	    zw.addEntryStream("OPS/" + it.name, my_modtime,
			      zw.COMPRESSION_DEFAULT, stream, false);
	}

	for (let it in loop(this.nonspine)) {
	    let my_modtime = modtime;
	    if (it.string)
		stream = converter.convertToInputStream(it.string);
	    else	// href
		stream = hrefToStream(it.href);
	    zw.addEntryStream("OPS/" + it.name, my_modtime,
			      zw.COMPRESSION_DEFAULT, stream, false);
	}

	debug("init: 10");

	zw.close();
	return zip_file;
    },

    generate_nav: function() {
 	let nav = '<html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">\n';
	nav += '<head>\n<title>' + escapeHtml(this.title) + '</title>\n';
	nav += '</head>\n<body>\n';
	nav += '<h1>' + escapeHtml(this.title) + '</h1>\n';
	nav += '<nav epub:type="toc" id="toc">\n';
	nav += '<ol>\n';
	for (let node in loop(this.nav)) {
	    nav += '<li>\n';
	    nav += '<a href="' + escapeHtml(node.href) + '">' + escapeHtml(node.title) + '</a>\n';
	    nav += '</li>\n';
	}
	nav += '</ol>\n';
	nav += '</nav>\n';
	nav += '</body>\n';
	nav += '</html>';
	return nav;
    },

    generate_titlepage: function() {
	let extra = null;
	let doc = '<?xml version="1.0" encoding="UTF-8"?>\n';
	doc += '<!DOCTYPE html>\n';
	doc += '<html xmlns="http://www.w3.org/1999/xhtml">\n';
	doc += '<head>\n';
	doc += '<title>' + escapeHtml(this.title) + '</title>\n';
	doc += '</head>\n';
	doc += '<body style="background-color: #ffffff; text-align: center; margin: 0px 0px 0px 0px;">\n';
	doc += '<div style="font-family: Gentium, \'URW Palladio L\', Georgia, serif; margin-bottom: 8em;">\n';
	doc += '<div style="color: #7f0000; font-size: 2.4em; margin-top: 4em; margin-bottom: 0.3em; margin-left: 1.5em; margin-right: 1.5em;">' + escapeHtml(this.title) + '</div>\n';

	let el;
	if (this.creators.length > 0) {
	    // FIXME
	    doc += '<div style="color: #00003f; font-size: 2.2em; font-variant: small-caps; margin-left: 1.5em; margin-right: 1.5em;">' + escapeHtml(this.creators[0].name) + '</div>\n';
	}

	doc += '</div>\n';

	if (this.publisher) {
	    doc += '<div style="color: #df0000; font-family: \'Doulos SIL\', \'MgOpen Canonica\', \'Times New Roman\', serif; font-size: 1.7em; margin-bottom: 2.5em;">' + escapeHtml(this.publisher) + '</div>\n';
	}

	doc += '</body>\n';
	doc += '</html>';

	return [doc, extra];
    },

    wrap_content: function(title, content) {
	let doc = '<?xml version="1.0" encoding="UTF-8"?>\n';
	doc += '<!DOCTYPE html>\n';
	doc += '<html xmlns="http://www.w3.org/1999/xhtml">\n';
	doc += '<head>\n';
	doc += '<title>' + escapeHtml(title) + '</title>\n';
	doc += '</head>\n';
	doc += '<body>\n';
	doc += content;
	doc += '</body>\n';
	doc += '</html>';
	return doc;
    },

    escapeHtml: function(text) {
	return escapeHtml(text);
    }
};

var components = [epubCreator];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

function hrefToStream(href) {
    let curi = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService).
	newURI(href, null, null);
    let uri = Components.classes["@mozilla.org/chrome/chrome-registry;1"].
	getService(Components.interfaces.nsIChromeRegistry).
	convertChromeURL(curi);
    let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	getService(Components.interfaces.nsIFileProtocolHandler).
	getFileFromURLSpec(uri.spec);
    let is = Components.classes["@mozilla.org/network/file-input-stream;1"].
	createInstance(Components.interfaces.nsIFileInputStream);
    is.init(file, -1, -1, 0);
    return is;
}

function escapeHtml(text) {
    text = text.replace(/&/g, '&amp;');
    text = text.replace(/</g, '&lt;');
    text = text.replace(/>/g, '&gt;');
    return text;
}
